.data
r1 dd ?
r2 dd ?
r3 dd ?
r4 dd ?
ip dd ?
baseip dd ?
stackp dd ?
stack dd ?
flags dd ?
.code
miniVm proc
	
	pop ebx
	pop eax
	mov [stack],eax
	pop eax
	mov [ip],eax
	mov [baseip],eax
	pusha
	mov [flags],0
	mov [stackp],0
	call _core
	popa
	push ebx
	ret

miniVm endp

_core proc
	
_next_ip:
	mov ebx,[ip]
	xor eax,eax
	xor ecx,ecx
	xor edx,edx
	mov al,byte ptr [ebx]		; al = instructional opcode
	mov dl,al
	and dl,0Fh		; Major opcode command
	and al,0F0h		; Minor opcode command
	cmp al,0C0h		; MOV r32
	je _call_mov
	cmp al,0B0h		; MOV r32 contents
	je _call_mov_off
	cmp al,0D0h		; CMP r32,value
	je _call_cmp
	cmp al,0E0h		; INC/DEC
	je _call_math
	cmp al, 0		; XOR
	je _call_xor
	cmp al, 10h		; JMP
	je _call_jmp
	cmp al, 20h		; Conditional JMP
	je _call_if_jmp
	cmp al, 30h		; PUSH/POP
	je _call_stack
	cmp al, 50h		; CALL
	je _call_call
	cmp al, 60h		; ADD/SUB
	je _call_add
	cmp al, 40h		; MISC
	jne _inc_ip
	cmp dl, 3
	je _exit_core
	jmp _call_misc
	; Unknown calls are treated like NOPs
_core_inc_ip:
	call _inc_ip
_core_next_ip:
	jmp _next_ip

_call_mov:
	call __mov
	jmp _core_inc_ip
_call_mov_off:
	call __mov_off
	jmp _core_inc_ip
_call_cmp:
	call __cmp
	jmp _core_inc_ip
_call_math:
	call __math
	jmp _core_inc_ip
_call_xor:
	call __xor
	jmp _core_inc_ip
_call_jmp:
	call __jmp
	jmp _core_next_ip
_call_if_jmp:
	call __if_jmp
	cmp dh,0
	je _call_if_jmp_inc
	jmp _core_next_ip
_call_if_jmp_inc:
	jmp _core_inc_ip
_call_stack:
	call __stack
	jmp _core_inc_ip
_call_call:
	call __call
	jmp _core_inc_ip
_call_add:
	call __add
	jmp _core_inc_ip
_call_misc:
	call __misc
	jmp _core_inc_ip
_exit_core:
	ret

_core endp

_inc_ip proc
	push ebx
	mov ebx,[ip]
	inc ebx
	mov [ip],ebx
	pop ebx
	ret

_inc_ip endp
_get_r32 proc
	
	push ebx
	mov ebx,[ip]
	inc ebx
	mov eax,[ebx]
	add ebx,3
	mov [ip],ebx
	pop ebx
	ret

_get_r32 endp
__mov proc
	call _get_r32
	cmp dl,7
	jg __mov_direct_value
	mov ecx,[ip]
	add eax,ecx				; EAX = IP + VALUE
	add dl,8				; Then adjust to consider direct value
__mov_direct_value:
	cmp dl,8
	je __mov_r1_v
	cmp dl,9
	je __mov_r2_v
	cmp dl,0Ah
	je __mov_r3_v
	cmp dl,0Bh
	je __mov_r4_v
	cmp dl,0Ch
	je __mov_ip_v
	cmp dl,0Dh
	je __mov_stack_v
	cmp dl,0Eh
	je __mov_flags_v
	jmp __mov_exit
__mov_r1_v:
	mov [r1],eax
	jmp __mov_exit
__mov_r2_v:
	mov [r2],eax
	jmp __mov_exit
__mov_r3_v:
	mov [r3],eax
	jmp __mov_exit
__mov_r4_v:
	mov [r4],eax
	jmp __mov_exit
__mov_ip_v:
	mov [ip],eax
	jmp __mov_exit
__mov_stack_v:
	mov [stack],eax
	jmp __mov_exit
__mov_flags_v:
	mov [flags],eax
__mov_exit:
	ret

__mov endp

__mov_off proc
	cmp dl,7
	jg __mov_to_v
	cmp dl,0
	je __mov_get_r2
	cmp dl,1
	je __mov_get_r3
	cmp dl,2
	je __mov_get_r4
	cmp dl,3
	je __mov_get_ip
	cmp dl,4
	je __mov_get_stack
	cmp dl,5
	je __mov_get_flags
	jmp __mov_off_exit
__mov_get_r2:
	mov eax,[r2]
	jmp _mov_to_r1
__mov_get_r3:
	mov eax,[r3]
	jmp _mov_to_r1
__mov_get_r4:
	mov eax,[r4]
	jmp _mov_to_r1
__mov_get_ip:
	mov eax,[ip]
	jmp _mov_to_r1
__mov_get_stack:
	mov eax,[stack]
	jmp _mov_to_r1
__mov_get_flags:
	mov eax,[flags]
	jmp _mov_to_r1
_mov_to_r1:
	mov ecx,offset r1
	jmp __mov_to_off
__mov_to_v:
	call _get_r32
	mov ecx,eax
	cmp dl,8
	je __mov_r1_to_v
	cmp dl,9
	je __mov_r2_to_v
	cmp dl,0Ah
	je __mov_r3_to_v
	cmp dl,0Bh
	je __mov_r4_to_v
	cmp dl,0Ch
	je __mov_ip_to_v
	cmp dl,0Dh
	je __mov_stack_to_v
	cmp dl,0Eh
	je __mov_flags_to_v
	jmp __mov_off_exit
__mov_r1_to_v:
	mov [r1],ecx
	jmp __mov_off_exit	
__mov_r2_to_v:
	mov [r2],ecx
	jmp __mov_off_exit	
__mov_r3_to_v:
	mov [r3],ecx
	jmp __mov_off_exit	
__mov_r4_to_v:
	mov [r4],ecx
	jmp __mov_off_exit	
__mov_ip_to_v:
	mov [ip],ecx
	jmp __mov_off_exit	
__mov_stack_to_v:
	mov [stack],ecx
	jmp __mov_off_exit	
__mov_flags_to_v:	
	mov [flags],ecx
	jmp __mov_off_exit	
__mov_to_off:
	mov [ecx],eax
	jmp __mov_off_exit
__mov_off_exit:
	ret

__mov_off endp
__cmp proc
	mov [flags],0
	cmp dl,7
	jg __cmp_to_v
	cmp dl,0
	je __cmp_r1_to_r2
	jmp __cmp_exit
__cmp_r1_to_r2:
	mov eax,[r1]
	mov edx,[r2]
	jmp __cmp_set_flags
__cmp_to_v:
	call _get_r32
	cmp dl,8
	je __cmp_r1_v
	cmp dl,9
	je __cmp_r2_v
	cmp dl,0Ah
	je __cmp_r3_v
	cmp dl,0Bh
	je __cmp_r4_v
	cmp dl,0Ch
	je __cmp_ip_v
	cmp dl,0Dh
	je __cmp_stack_v
	cmp dl,0Eh
	je __cmp_flags_v
	jmp __cmp_exit
__cmp_r1_v:
	mov edx,[r1]
	jmp __cmp_set_flags
__cmp_r2_v:
	mov edx,[r2]
	jmp __cmp_set_flags
__cmp_r3_v:
	mov edx,[r3]
	jmp __cmp_set_flags
__cmp_r4_v:
	mov edx,[r4]
	jmp __cmp_set_flags
__cmp_ip_v:
	mov edx,[ip]
	jmp __cmp_set_flags
__cmp_stack_v:
	mov edx,[stack]
	jmp __cmp_set_flags
__cmp_flags_v:
	mov edx,[flags]
__cmp_set_flags:
	cmp edx,eax
	mov edx,[flags]
	je __cmp_set_flags_true
	and edx,1
	jl __cmp_set_flags_lt
	or edx,2
	mov [flags],edx
	jmp __cmp_exit
__cmp_set_flags_lt:
	and edx,2
	mov [flags],edx
	jmp __cmp_exit
__cmp_set_flags_true:
	or edx,1
	mov [flags],edx
__cmp_exit:
	ret

__cmp endp
__math proc
	cmp dl,0
	je __dec_r1
	cmp dl,1
	je __dec_r2
	cmp dl,2
	je __dec_r3
	cmp dl,3
	je __dec_r4
	cmp dl,4
	je __dec_ip
	cmp dl,5
	je __dec_stack
	cmp dl,8
	je __inc_r1
	cmp dl,9
	je __inc_r2
	cmp dl,0Ah
	je __inc_r3
	cmp dl,0Bh
	je __inc_r4
	cmp dl,0Ch
	je __inc_ip
	cmp dl,0Dh
	je __inc_stack
	jmp __math_exit
__dec_r1:
	mov eax,[r1]
	dec eax
	mov [r1],eax
	jmp __math_exit
__dec_r2:
	mov eax,[r2]
	dec eax
	mov [r2],eax
	jmp __math_exit
__dec_r3:
	mov eax,[r3]
	dec eax
	mov [r3],eax
	jmp __math_exit
__dec_r4:
	mov eax,[r4]
	dec eax
	mov [r4],eax
	jmp __math_exit
__dec_ip:
	mov eax,[ip]
	dec eax
	mov [ip],eax
	jmp __math_exit
__dec_stack:
	mov eax,[stack]
	dec eax
	mov [stack],eax
	jmp __math_exit
__inc_r1:
	mov eax,[r1]
	inc eax
	mov [r1],eax
	jmp __math_exit
__inc_r2:
	mov eax,[r2]
	inc eax
	mov [r2],eax
	jmp __math_exit
__inc_r3:
	mov eax,[r3]
	inc eax
	mov [r3],eax
	jmp __math_exit
__inc_r4:
	mov eax,[r4]
	inc eax
	mov [r4],eax
	jmp __math_exit
__inc_ip:
	mov eax,[ip]
	inc eax
	mov [ip],eax
	jmp __math_exit
__inc_stack:
	mov eax,[stack]
	inc eax
	mov [stack],eax
__math_exit:
	ret

__math endp

__xor proc
	mov eax,[r1]
	cmp dl,0
	je __xor_r1
	cmp dl,1
	je __xor_r2
	cmp dl,2
	je __xor_r3
	cmp dl,3
	je __xor_r4
	cmp dl,4
	je __xor_ip
	cmp dl,5
	je __xor_stack
	jmp __xor_exit
__xor_r1:
	mov edx,0
	jmp __xor_set_r1
__xor_r2:
	mov edx,[r2]
	xor edx,eax
	jmp __xor_set_r1
__xor_r3:
	mov edx,[r3]
	xor edx,eax
	jmp __xor_set_r1
__xor_r4:
	mov edx,[r4]
	xor edx,eax
	jmp __xor_set_r1
__xor_ip:
	mov edx,[ip]
	xor edx,eax
	jmp __xor_set_r1
__xor_stack:
	mov edx,[stack]
	xor edx,eax
__xor_set_r1:
	mov [r1],edx
__xor_exit:
	ret

__xor endp
__jmp proc
	call _get_r32
	mov ebx,[baseip]
	cmp dl,0
	je __jmp_relative_v
	cmp dl,1
	je __jmp_direct_v
	jmp __jmp_exit
__jmp_relative_v:
	add eax,ebx
	mov [ip],eax
	jmp __jmp_exit
__jmp_direct_v:
	mov [ip],eax
__jmp_exit:
	ret

__jmp endp
__if_jmp proc
	call _get_r32
	mov ebx,[baseip]
	add ebx, eax
	mov eax,[flags]
	mov dh, 0	; DH = did we jump flag
	cmp dl,0
	je __if_jmp_eq_v
	cmp dl,1
	je __if_jmp_lt_v
	cmp dl,2
	je __if_jmp_gt_v
	jmp __if_jmp_exit
__if_jmp_eq_v:
	and eax,1
	jmp __jmp_match
__if_jmp_lt_v:
	and eax,2
	jmp __jmp_lt_match
__if_jmp_gt_v:
	and eax,2
	jmp __jmp_gt_match
__jmp_match:
	cmp al,1
	jne __if_jmp_exit
	mov [ip],ebx
	mov dh,1
	jmp __if_jmp_exit
__jmp_lt_match:
	cmp al,2
	je __if_jmp_exit
	mov [ip],ebx
	mov dh,1
	jmp __if_jmp_exit
__jmp_gt_match:
	cmp al,2
	jne __if_jmp_exit
	mov [ip],ebx
	mov dh,1
__if_jmp_exit:
	ret

__if_jmp endp

__stack proc
	mov ecx,[stackp]
	mov ebx,[stack]
	add ebx,ecx			; EBX = Current Stack Location
	cmp dl,0
	je __push_r1
	cmp dl,1
	je __push_r2
	cmp dl,2
	je __push_r3
	cmp dl,3
	je __push_r4
	cmp dl,4
	je __push_ip
	cmp dl,5
	je __push_stack
	cmp dl,6
	je __push_flags
	cmp dl,8
	je __pop_r1
	cmp dl,9
	je __pop_r2
	cmp dl,0Ah
	je __pop_r3
	cmp dl,0Bh
	je __pop_r4
	cmp dl,0Ch
	je __pop_ip
	cmp dl,0Dh
	je __pop_stack
	cmp dl,0Eh
	je __pop_flags
	jmp __stack_exit
__push_r1:
	mov eax,[r1]
	jmp __stack_inc
__push_r2:
	mov eax,[r2]
	jmp __stack_inc
__push_r3:
	mov eax,[r3]
	jmp __stack_inc
__push_r4:
	mov eax,[r4]
	jmp __stack_inc
__push_ip:
	mov eax,[ip]
	jmp __stack_inc
__push_stack:
	mov eax,[stackp]
	jmp __stack_inc
__push_flags:
	mov eax,[flags]
	jmp __stack_inc
__pop_r1:
	mov eax,[ebx]
	mov [r1],eax
	jmp __stack_dec
__pop_r2:
	mov eax,[ebx]
	mov [r2],eax
	jmp __stack_dec
__pop_r3:
	mov eax,[ebx]
	mov [r3],eax
	jmp __stack_dec
__pop_r4:
	mov eax,[ebx]
	mov [r4],eax
	jmp __stack_dec
__pop_ip:
	mov eax,[ebx]
	mov [ip],eax
	jmp __stack_dec
__pop_stack:
	mov eax,[ebx]
	mov [stackp],eax
	jmp __stack_dec
__pop_flags:
	mov eax,[ebx]
	mov [flags],eax
	jmp __stack_dec
__stack_inc:
	sub ebx,4
	mov [ebx],eax
	sub ecx,4
	mov [stackp],ecx
	jmp __stack_exit
__stack_dec:
	add ecx,4
	mov [stackp],ecx
__stack_exit:
	ret

__stack endp
__misc proc
	cmp dl,0
	je __misc_and_r1
	cmp dl,1
	je __misc_or_r1
	cmp dl,2
	je __misc_mov_r1_r1
	jmp __misc_exit
__misc_and_r1:
	call _get_r32
	mov edx,[r1]
	and edx,eax
	mov [r1],edx
	jmp __misc_exit	
__misc_or_r1:
	call _get_r32
	mov edx,[r1]
	or edx,eax
	mov [r1],edx
	jmp __misc_exit
__misc_mov_r1_r1:
	mov eax,[r1]
	mov eax,[eax]
	mov [r1],eax
	jmp __misc_exit
__misc_exit:
	ret
__misc endp
__add proc
	
	cmp dl, 7
	jg __sub_r1
	cmp dl, 0
	je __add_r1_r1
	cmp dl, 1
	je __add_r1_r2
	cmp dl, 2
	je __add_r1_r3
	cmp dl, 3
	je __add_r1_r4
	cmp dl, 4
	je __add_r1_ip
	cmp dl, 5
	je __add_r1_stack
	jmp __add_exit
__sub_r1:
	cmp dl, 8
	je __sub_r1_r1
	cmp dl, 9
	je __sub_r1_r2
	cmp dl, 10
	je __sub_r1_r3
	cmp dl, 11
	je __sub_r1_r4
	cmp dl, 12
	je __sub_r1_ip
	cmp dl, 13
	je __sub_r1_stack
	jmp __add_exit
__add_r1_r1:
    mov eax, [r1]
    jmp __add_to_r1
__add_r1_r2:
    mov eax, [r2]
    jmp __add_to_r1
__add_r1_r3:
    mov eax, [r3]
    jmp __add_to_r1
__add_r1_r4:
    mov eax, [r4]
    jmp __add_to_r1
__add_r1_ip:
    mov eax, [ip]
    jmp __add_to_r1
__add_r1_stack:
    mov eax, [stackp]
    jmp __add_to_r1
__add_to_r1:
	mov ebx, [r1]
	add ebx,eax
	mov [r1],ebx
__add_exit:	
	ret
__sub_r1_r1:
    mov eax, [r1]
    jmp __sub_to_r1
__sub_r1_r2:
    mov eax, [r2]
    jmp __sub_to_r1
__sub_r1_r3:
    mov eax, [r3]
    jmp __sub_to_r1
__sub_r1_r4:
    mov eax, [r4]
    jmp __sub_to_r1
__sub_r1_ip:
    mov eax, [ip]
    jmp __sub_to_r1
__sub_r1_stack:
    mov eax, [stackp]
    jmp __sub_to_r1
__sub_to_r1:
	mov ebx, [r1]
	sub ebx,eax
	mov [r1],ebx
__sub_exit:
	ret
__add endp
__call proc
	
	cmp dl,7
	jg __call_v
	cmp dl,0
	je __call_r1
__call_r1:
	mov eax,[r1]
	call eax
__call_v:
	call _get_r32
	call eax
	jmp __call_exit
__call_exit:
	ret

__call endp